static const char _rcsid_[]="$Id: xiosigchld.c,v 1.11 2001/09/16 07:55:03 gerhard Exp $";
/* Copyright Gerhard Rieger 2001 */
/* Published under the GNU General Public License V.2, see file COPYING */

/* this is the source of the extended child signal handler */


#include "xiosysincludes.h"
#include "xioopen.h"


/*! */
extern int closing;
extern int sleepon;

/*! with socat, at most 4 exec children exist */
pid_t diedunknown1;	/* child died before it is registered */
pid_t diedunknown2;
pid_t diedunknown3;
pid_t diedunknown4;


static int sigchld_stream(struct single *file) {
   file->para.exec.pid = 0;
#if 1
   if (file->ignoreeof && !closing) {
      sleepon = 1;
   } else {
      file->eof = 1; closing = 1;
   }
#else
   file->eof = 1;
#endif
   return 0;
}

/* return 0 if socket is not responsible for deadchild */
static int xio_checkchild(xiofile_t *socket, int socknum, pid_t deadchild) {
   if (socket != NULL) {
      if (socket->tag != XIO_TAG_DUAL) {
	 if ((socket->stream.howtoend == END_KILL ||
	      socket->stream.howtoend == END_CLOSE_KILL ||
	      socket->stream.howtoend == END_SHUTDOWN_KILL) &&
	     socket->stream.para.exec.pid == deadchild) {
	    Info2("exec'd process %d on socket %d terminated",
		  socket->stream.para.exec.pid, socknum);
	    sigchld_stream(&socket->stream);
	    return 1;
	 }
      } else {
	 if ((socket->dual.stream[0]->stream.howtoend == END_KILL ||
	      socket->dual.stream[0]->stream.howtoend == END_CLOSE_KILL ||
	      socket->dual.stream[0]->stream.howtoend == END_SHUTDOWN_KILL) &&
	     socket->dual.stream[0]->stream.para.exec.pid == deadchild) {
	    Info1("exec'd process %d on socket-input terminated",
		  socket->dual.stream[0]->stream.para.exec.pid);
	    sigchld_stream(&socket->dual.stream[0]->stream);
	    return 1;
	 } else if ((socket->dual.stream[1]->stream.howtoend == END_KILL ||
		     socket->dual.stream[1]->stream.howtoend == END_CLOSE_KILL ||
		     socket->dual.stream[1]->stream.howtoend == END_SHUTDOWN_KILL) &&
		    socket->dual.stream[1]->stream.para.exec.pid == deadchild) {
	    Info1("exec'd process %d on socket-output terminated",
		  socket->dual.stream[1]->stream.para.exec.pid);
	    sigchld_stream(&socket->dual.stream[1]->stream);
	    return 1;
	 }
      }
   }
   return 0;
}

/* this is the signal handler for SIGCHLD */
/* the current socat/xio implementation knows two kinds of children:
   exec/system addresses perform a fork: these children are registered and
   there death influences the parents flow;
   listen-socket with fork children: these children are "anonymous" and their
   death does not effect the parent process (now; maybe we have a child
   process counter later) */
void childdied(int signum) {
   pid_t pid;
   int status;

   Info1("childdied(signum=%d)", signum);
   if ((pid = Waitpid(-1, &status, WNOHANG)) < 0) {
      Warn2("waitpid(-1, {%d}, WNOHANG): %s", status, strerror(errno));
   }
   if (!xio_checkchild(sock1, 1, pid)) {
      if (!xio_checkchild(sock2, 2, pid)) {
	 Info2("childdied(%d): cannot identify child %d", signum, pid);
	 if (diedunknown1 == 0) diedunknown1 = pid;
	 else if (diedunknown2 == 0) diedunknown2 = pid;
	 else if (diedunknown3 == 0) diedunknown3 = pid;
	 else diedunknown4 = pid;
      }
   }

   if (WIFEXITED(status)) {
      if (WEXITSTATUS(status) == 0) {
	 Info3("waitpid(-1, {%d}, WNOHANG): child %d exited with status %d",
	       status, pid, WEXITSTATUS(status));
      } else {
	 Warn3("waitpid(-1, {%d}, WNOHANG): child %d exited with status %d",
		status, pid, WEXITSTATUS(status));
      }
   } else if (WIFSIGNALED(status)) {
      Info3("waitpid(-1, {%d}, WNOHANG): child %d exited on signal %d",
	    status, pid, WTERMSIG(status));
   } else if (WIFSTOPPED(status)) {
      Info3("waitpid(-1, {%d}, WNOHANG): child %d is stopped on signal %d",
	    status, pid, WSTOPSIG(status));
   } else {
      Warn2("waitpid(-1, {%d}, WNOHANG): cannot determine status of child %d",
	     status, pid);
   }

   if (Signal(SIGCHLD, childdied) == SIG_ERR) {
      Warn2("signal(SIGCHLD, %p): %s", childdied, strerror(errno));
   }
   Info("childdied() finished");
}
